"""
Copyright (c) 2022 Center for Astrophysics | Harvard & Smithsonian

This software is licensed for use as described in the LICENSE file in
the root directory of this distribution.

Code to manage targets

Originator: Aaron Oppenheimer March 2020
"""
from pathlib import Path
import csv
import pandas as pd
import numpy as np

THE_TARGETS = None

class Target:
    name = None

    @staticmethod
    def get_list():
        return list(THE_TARGETS.keys())

    @staticmethod
    def from_name(name):
        return THE_TARGETS[name]

    @classmethod
    def get_default(cls):
        return cls.from_name(cls.get_list()[0])

    def __init__(self, name, **kwargs):
        self.name = name
        for k,v in kwargs.items():
            setattr(self, k, v)

    def __str__(self):
        return f'Target {self.name}'

    def __repr__(self):
        return f'Target {self.name}'


 ## Helper functions

def _init_targets():
    """ do the initial setup on arrays """
    global THE_TARGETS
    if THE_TARGETS is None:
        path=str(Path(__file__).parent) + '/config'
        targs = pd.read_csv(f'{path}/targets.csv', index_col=0)

        def convert_ra_dec(site):
            ra = site['RA_hr']+site['RA_min']/60.0+site['RA_sec']/3600.
            dec = (np.sign(site['Dec_deg'])*(np.abs(site['Dec_deg'])+site['Dec_arcmin']/60.0+site['Dec_arcsec']/3600.))
            return [ra, dec]

        targs[['RA','Dec']] = targs.apply(convert_ra_dec, axis=1, result_type='expand')
        THE_TARGETS = {x:Target(name=x, **(targs.loc[x].to_dict())) for x in targs.index}

_init_targets()



"""
Copyright (c) 2022 Center for Astrophysics | Harvard & Smithsonian

This software is licensed for use as described in the LICENSE file in
the root directory of this distribution.

Plots a map with antenna locations shown.

Requires the plotly library ('pip install plotly')

Originator: Aaron Oppenheimer March 2020
"""

import plotly.graph_objs as go
from ngehtutil.station import Station


eht_sites = [
    'ALMA', 
    'APEX', 
    'GLT', 
    'IRAM', 
    'JCMT', 
    'KP', 
    'LMT', 
    'NOEMA', 
    'SMA', 
    'SMT', 
    'SPT'
]

phase_1_sites = [
    'HAY', 
    'OVRO', 
    'BAJA', 
    'CNI', 
    'LAS'
]

phase_2_sites = [
    'BAN', 
    'BAR', 
    'BGA', 
    'BMAC', 
    'BOL', 
    'BRZ', 
    'CAS', 
    'CAT', 
    'FAIR', 
    'GAM', 
    'GARS', 
    'GLTS', 
    'JELM', 
    'KILI', 
    'LLA', 
    'PAR', 
    'PIKE', 
    'SAN', 
    'SGO', 
    'SKS', 
    'SPX', 
    'TRL', 
    'YAN'
]

sites=[eht_sites, phase_1_sites, phase_2_sites]
legend=['Existing EHT', 'ngEHT Phase 1', 'Poss. ngEHT Phase 2']
colors=['blue','orange', 'white']

centerlon = -100

show_names = False

fig = go.Figure()
for i,lst in enumerate(sites):
    lats = []
    lons = []
    names = []
    for s in lst:
        s=Station.from_name(s)
        lats.append(s.latitude)
        lons.append(s.longitude)
        names.append(s.name)

    fig.add_trace(
        go.Scattergeo(
            lon = lons,
            lat = lats,
            text = names if show_names else None,
            textposition="top center",
            mode = 'markers+text',
            marker = dict(
                size = 10,
                color = colors[i],
                line = dict(
                    width = 1,
                    color = 'black'
                )
            ),
            name = legend[i]
        )
    )

fig.update_geos(
    center=dict(lat=0, lon=-100), # this will center on the point
    lataxis=dict(range=[-90,90]),
    lonaxis=dict(range=[-180+centerlon,180+centerlon]),
)

fig.update_layout(legend=dict(
    yanchor="bottom",
    y=0.05,
    xanchor="left",
    x=0.01,
    borderwidth = 1,
))

fig.show()



"""
Copyright (c) 2022 Center for Astrophysics | Harvard & Smithsonian

This software is licensed for use as described in the LICENSE file in
the root directory of this distribution.

Cost model code

Originator: Aaron Oppenheimer March 2020
"""
from math import floor
import pandas as pd
import logging

def calculate_costs_over_time(cost_config, array_costs):
    """ Do the calculation of array costs for each of the arrays described in array_data """
    costs_over_time = pd.DataFrame(dtype=float)

    years_advance_spend = 3 # years before we break ground that we start spending capex
    calculation_start = 2020

    logging.info(f'calculating costs over time')

    years_to_build = cost_config.fully_operational - cost_config.start_building + years_advance_spend
    year_first_build = cost_config.start_building - years_advance_spend
    final_year = cost_config.fully_operational + cost_config.active_lifetime
    inflation = 1 + cost_config.inflation_rate
    average_annual_capex = array_costs.loc['TOTAL CAPEX'] / years_to_build
    average_annual_opex = array_costs.loc['ANNUAL OPEX'] / array_costs.loc['Total Sites Count']

    existing_stations = array_costs.loc['EHT Sites Count']
    stations_to_build = array_costs.loc['New Sites Count']
    stations_built_per_year = stations_to_build / years_to_build
    stations_built = 0
    for year in range(calculation_start, final_year+1):
        building_in_progress = year >= year_first_build and year <= cost_config.fully_operational
        stations_operating = year >= cost_config.start_building

        cap_spend = (average_annual_capex * pow(inflation, year-calculation_start)) \
            if building_in_progress else 0

        stations_built += stations_built_per_year \
            if building_in_progress and year < cost_config.fully_operational else 0

        stations_available = existing_stations + floor(stations_built)

        op_spend = (stations_available * average_annual_opex * \
                    pow(inflation, year - calculation_start)) \
                        if stations_operating else 0

        costs_over_time.loc[year,'CAPEX'] = float(cap_spend)
        costs_over_time.loc[year,'OPEX'] = float(op_spend)

    return costs_over_time



"""
Copyright (c) 2022 Center for Astrophysics | Harvard & Smithsonian

This software is licensed for use as described in the LICENSE file in
the root directory of this distribution.

Cost model code

Originator: Aaron Oppenheimer March 2020
"""

class CostConfig:
    dish_size = 6 # Size for new dishes - must be 4, 6, 8, 10
    autonomy_of_operations = 'Manual' # one of 'Manual' 'Somewhat Autonomous' 'Semi-Autonomous'
                                      # or 'Fully Autonomous'
    data_management = 'Own Cluster' # one of 'Own Cluster' 'Research Cluster' 'Cloud'
    recording_bandwidth = 8 # in GHz
    recording_frequencies = 2
    start_building = 2025
    fully_operational = 2030
    inflation_rate = 0.02
    active_lifetime = 10
    observations_per_year = 1
    days_per_observation = 3
    hours_per_observation = 30
    no_upgrade = [] # sites to skip when calculating costs for receiver, backend, etc.

    def __init__(self, **kwargs):
        """
        Pass in values to override defaults
        """
        for k,v in kwargs.items():
            if not hasattr(self, k):
                raise KeyError(f'no configuration key "{k}"')
            setattr(self, k, v)



"""
Copyright (c) 2022 Center for Astrophysics | Harvard & Smithsonian

This software is licensed for use as described in the LICENSE file in
the root directory of this distribution.

Code to demonstrate use of the cost model

Originator: Aaron Oppenheimer March 2020
"""
from ngehtutil.cost import calculate_costs, CostConfig
from ngehtutil import *


def main():

    config = CostConfig(
        observations_per_year=1,
        days_per_observation=60,
        hours_per_observation=60*8,
        no_upgrade = ['ALMA', 'NOEMA', 'SMA']
    )
    array = Array.from_name('ngEHT Ref. Array 1.1A + EHT2022')
    costs, _ = calculate_costs(config, array.stations())
    
    print(f'array: {array}')
    print('CONFIGURATION:')
    for x in dir(config):
        if not x[0]=='_':
            print(f'{x}: {getattr(config,x)}')
    for k,v in costs.items():
        print(f'{k}: {v}')


if __name__ == '__main__':
    main()



"""
Copyright (c) 2022 Center for Astrophysics | Harvard & Smithsonian

This software is licensed for use as described in the LICENSE file in
the root directory of this distribution.

Cost model code

Originator: Aaron Oppenheimer March 2020
"""
import functools
import operator
import logging
import os
import pandas as pd
from ngehtutil.station import Station
from ngehtutil.version import VERSION
import math

# constants from file - we don't want to have to read these every time
CONSTANTS_TABLES = None

def read_constants(infile_name=None):
    """ Read in all of the constants used for calculations as pandas dataframes """
    global CONSTANTS_TABLES # pylint: disable=global-statement
    if CONSTANTS_TABLES is None:

        if infile_name is None:
            infile_name = f'{os.path.dirname(__file__)}/cost_constants.xlsx'

        logging.info('loading cost contants from %s', infile_name)

        tables = {}
        tables['site_development_values_table'] = pd.read_excel(
            infile_name, sheet_name="SiteDevelopmentValues", index_col=0)
        tables['labor_cost_values_table'] = pd.read_excel(
            infile_name, sheet_name="LaborCostValues", index_col=0)
        tables['travel_cost_values_table'] = pd.read_excel(
            infile_name, sheet_name="TravelCostValues", index_col=0)
        tables['autonomy_mode_values_table'] = pd.read_excel(
            infile_name, sheet_name="AutonomyModeValues", index_col=0)
        tables['data_management_values_table'] = pd.read_excel(
            infile_name, sheet_name="DataManagementValues", index_col=0)
        tables['data_management_option_values_table'] = pd.read_excel(
            infile_name, sheet_name="DataManagementOptionValues", index_col=0)
        CONSTANTS_TABLES = tables

    return CONSTANTS_TABLES

def get_cost_constants():
    """
    for purposes of processing on a cluster, we need a deep copy of the constants that
    can be passed to the cost calculation functions instead of using the global one which
    we think is causing headaches.
    """
    if CONSTANTS_TABLES is None:
        read_constants()
    new_const = {k:v.copy() for k,v in CONSTANTS_TABLES.items()}
    return new_const


def calculate_costs(cost_config, sites, cost_constants=None, const_filename=None):
    """
    Perform cost calculations for an array of ngEGT sites

    INPUTS
    ------
    cost_config: CostConfig object

    sites: list of Station objects

    OUTPUTS
    -------
    dataframe of cost information
    """
    # Read the cost constant spreadsheet
    if cost_constants:
        const = cost_constants
    else:
        const = read_constants(const_filename)

    # const = update_constants_from_config(const, cost_config)

    logging.info('calculating costs')

    #
    # Calculate statistics like number of sites, collecting days, amount of data collected
    #
    array_stats = {}
    # add a row just to hold the name of the category.
    array_stats['ARRAY STATS'] = VERSION

    total_sites_count = len(sites)
    # new_sites = sites.loc[:,sites.loc['eht']==0].to_dict()
    new_sites = [x for x in sites if x.dishes is None]
    new_sites_count = len(new_sites)
    array_stats['New Sites Count'] = new_sites_count
    array_stats['EHT Sites Count'] = total_sites_count - new_sites_count
    array_stats['Total Sites Count'] = total_sites_count

    array_stats = pd.Series(array_stats)
    array_stats = pd.concat([array_stats,calculate_operating_mode(cost_config, sites, const)])

    design_NRE = calculate_design_NRE(cost_config, sites, const)

    total_site_costs, new_site_costs, per_site_cap_costs = calculate_capital_costs(cost_config, sites, const)

    t, n, per_site_ops_costs = calculate_operations_costs(cost_config, sites, const)
    total_site_costs = pd.concat([total_site_costs, t])
    new_site_costs = pd.concat([new_site_costs, n])

    data_management_costs = calculate_data_costs(cost_config, total_sites_count, \
        array_stats["Data Per Year - Full Array (PB)"], \
            cost_config.observations_per_year*cost_config.days_per_observation, \
                const )

    #
    # work out the average costs for new sites
    #
    avg_new_site_build_costs = new_site_costs[1:].copy()
    new_names = {i:f'New Site Avg {i}' for i in avg_new_site_build_costs.index}
    avg_new_site_build_costs.rename(new_names, inplace=True)
    avg_new_site_build_costs[:] = (avg_new_site_build_costs[:] / new_sites_count) \
        if new_sites_count else 0
    avg_new_site_data_costs = data_management_costs[['Site Recorders','Site Media']]
    new_names = {i:f'New Site Avg {i}' for i in avg_new_site_data_costs.index}
    avg_new_site_data_costs.rename(new_names, inplace=True)
    avg_new_site_data_costs[:] = (avg_new_site_data_costs[:]/ total_sites_count) \
        if new_sites_count else 0

    # add a row just to hold the name of the category.
    avg_new_site_costs = pd.Series(dtype='float')
    avg_new_site_costs['NEW SITE AVG COSTS'] = ''
    avg_new_site_costs = pd.concat([avg_new_site_costs, \
                                    avg_new_site_build_costs, \
                                    avg_new_site_data_costs])

    capex_costs = [
        'New Site Avg Site acquisition / leasing',
        'New Site Avg Infrastructure',
        'New Site Avg Antenna construction',
        'New Site Avg Antenna transport',
        'New Site Avg Antenna commissioning',
        'New Site Avg Site Recorders',
        'New Site Avg Site Media',
    ]
    avg_new_site_costs['New Site Total CAPEX'] = \
        sum([avg_new_site_costs[i] for i in capex_costs])

    total_costs = pd.Series(dtype='float')
    total_costs['TOTAL COSTS'] = ''
    total_costs['DESIGN NRE'] = design_NRE
    total_costs['TOTAL CAPEX'] = total_site_costs[1:].drop('Antenna operations').sum() +\
                                 data_management_costs[['Cluster Build Cost',
                                                        'Site Recorders',
                                                        'Site Media']].sum()
    total_costs['ANNUAL OPEX'] = total_site_costs[['Antenna operations', 'Fulltime staff']].sum() +\
                                 data_management_costs[['Personnel',
                                                        'Holding Data Storage Costs',
                                                        'Fast Data Storage Costs',
                                                        'Transfer Costs',
                                                        'Computation Costs',
                                                        'Data Shipping']].sum()

    if total_sites_count:
        everything_per_site = pd.concat([per_site_cap_costs, per_site_ops_costs])
        everything_per_site.at['Data mgmt capital',:] = data_management_costs[['Site Recorders',
                                                                'Site Media']].sum() / total_sites_count
        everything_per_site.at['Data mgmt operations',:] = data_management_costs[['Personnel',
                                                            'Holding Data Storage Costs',
                                                            'Fast Data Storage Costs',
                                                            'Transfer Costs',
                                                            'Computation Costs',
                                                            'Data Shipping']].sum() / total_sites_count
        everything_per_site = everything_per_site.to_dict()
    else:
        everything_per_site = {}

    everything = pd.concat([pd.Series(array_stats), total_site_costs, data_management_costs, \
                      avg_new_site_costs, total_costs])
    return everything.to_dict(), everything_per_site


##
## OPERATION MODE CALCULATIONS
##

def calculate_operating_mode(cost_config, sites, const):
    array_stats = {}

    # calculate stuff about operation mode
    observations_per_year = cost_config.observations_per_year
    days_per_observation = cost_config.days_per_observation
    collecting_days_per_year = observations_per_year * days_per_observation
    hours_per_observation = cost_config.hours_per_observation
    collecting_hours_per_year = observations_per_year * hours_per_observation


    # calculate the amount of data collected per site per hour
    # 'bandwidth' can be overridden by specifying 'recording_bandwidth' in configuration
    vals = const['data_management_values_table'].loc[['bit_depth',
                                                  'sidebands', 'polarizations',
                                                  'oversampling']].Value
    bandwidth = cost_config.recording_bandwidth
    frequencies = cost_config.recording_frequencies
    gbits_per_second = functools.reduce(operator.mul, vals) * bandwidth * frequencies
    pb_per_hour = gbits_per_second / 8 * 3600 / 1e6
    pb_per_day = pb_per_hour * (hours_per_observation / days_per_observation)
    array_stats["Data Per Observation Per Station"] = pb_per_hour * hours_per_observation

    # calculate the total amount of data for this configuration
    total_data_collected_per_year = math.ceil(len(sites) * collecting_hours_per_year * pb_per_hour)
    array_stats["Data Per Year - Full Array (PB)"] = total_data_collected_per_year

    return pd.Series(array_stats)


##
## DESIGN NRE CALCULATIONS
##
def calculate_design_NRE(cost_config, sites, const):
    # calculate the total NRE costs for designing new stations
    nre_costs = 0
    if len(sites):
        # do we need to design a new antenna?
        if sum([0 if x.dishes else 1 for x in sites]):
            nre_costs = const['site_development_values_table'].at['antenna_development_nre','Value']
    return nre_costs


###
### CAPITAL COSTS FOR ANTENNAS
###
def calculate_capital_costs(cost_config, sites, const):

    #
    #  calculate costs based on the array configuration - NRE, construction costs, operating costs
    #
    cost_list = [
        'Site acquisition / leasing',
        'Infrastructure',
        'Support equipment',
        'Antenna construction',
        'Antenna transport',
        'Receiver and Backend costs',
        'Antenna commissioning',
    ]

    site_costs = pd.DataFrame(index=cost_list)
    total_site_costs = pd.Series(dtype='float') # holds total cost for all sites
    new_site_costs = pd.Series(dtype='float') # holds total cost for just new sites
    # add a row just to hold the name of the category.
    total_site_costs['SITE COSTS'] = ''
    new_site_costs['NEW SITE COSTS'] = ''

    # Antenna Construction
    #
    # implements the cost equation as C + kd^e
    dish_size = cost_config.dish_size
    antenna_constant = \
        const['site_development_values_table'].at['antenna_cost_constant','Value']
    antenna_factor1 = const['site_development_values_table'].at['antenna_cost_factor1','Value']
    antenna_factor2 = const['site_development_values_table'].at['antenna_cost_factor2','Value']
    antenna_exp = const['site_development_values_table'].at['antenna_cost_exponent','Value']
    single_antenna_cost = antenna_constant + \
                (antenna_factor1 * dish_size) + \
                (antenna_factor2 * pow(dish_size, antenna_exp))

    # now some numbers for each site, depending on its location, whether it already exists, etc.

    for site in sites:

        siteindex = site.name

        # set up the costs. Put in a junk value to make sure everything gets set properly
        site_costs.at[:, siteindex] = [-99] * len(cost_list)

        # For new sites we have to worry about costs to acquire, build, commission

        if site.dishes is None: # this is a new site

            # Aquisition Costs
            site_aquisition_baseline = const['site_development_values_table']\
                .at['site_acquisition_and_leasing', 'Value']
            site_costs.at['Site acquisition / leasing',
                            siteindex] = site_aquisition_baseline if site.site_acquisition else 0

            # Infrastructure Development
            infrastructure_baseline = const['site_development_values_table'].loc[
                'infrastructure_development', 'Value']
            infrascruture_scaling_factor = const['site_development_values_table'].loc[
                site.existing_infrastructure, 'Value']
            site_costs.at['Infrastructure',
                            siteindex] = infrastructure_baseline * infrascruture_scaling_factor


            # Support Equipment
            site_costs.at['Support equipment', siteindex] = \
                const['site_development_values_table'].loc['ground_support_equipment', 'Value']


            # site.set_diameter(dish_size) # the site was under-defined, so give it a dish

            construction_cost = single_antenna_cost

            construction_cost = construction_cost * \
                const['site_development_values_table']\
                    .at[site.polar_nonpolar, 'Value']  # polar multiplier

            site_costs.at['Antenna construction', siteindex] = construction_cost

            transport_cost = const['site_development_values_table'].loc[
                'antenna_transport_cost', 'Value']
            site_costs.at['Antenna transport', siteindex] = transport_cost

        else:
            # for existing sites, don't need to build dish
            site_costs.at['Site acquisition / leasing', siteindex] = 0
            site_costs.at['Infrastructure', siteindex] = 0
            site_costs.at['Support equipment', siteindex] = 0
            site_costs.at['Antenna construction', siteindex] = 0
            site_costs.at['Antenna transport', siteindex] = 0

        # Backend - receiver, maser
        if not site.name in cost_config.no_upgrade:
            receiver_cost = const['site_development_values_table']\
                .at['receiver_cost_factor', 'Value']
            if cost_config.recording_frequencies > 2:
                triband_cost_multiplier = const['site_development_values_table']\
                    .at['triband_cost_multiplier', 'Value']
                receiver_cost = receiver_cost * triband_cost_multiplier
            maser_cost = const['site_development_values_table'].at['maser_cost', 'Value']
            dbe_cost = const['site_development_values_table'].at['dbe_cost', 'Value'] * \
                cost_config.recording_frequencies
            num_dishes = len(site.dishes) if site.dishes else 1
            # add up costs and add a little for spare parts
            fe_be_cost = (receiver_cost + maser_cost + dbe_cost + (dbe_cost*.2)) * num_dishes
        else:
            fe_be_cost = 0
        site_costs.at['Receiver and Backend costs', siteindex] = fe_be_cost


        # Antenna Commissioning
        if site.dishes is None:
            site_costs.at['Antenna commissioning', siteindex] = \
                const['site_development_values_table']\
                    .at['commissioning_new', 'Value']
        else:
            site_costs.at['Antenna commissioning', siteindex] = \
                const['site_development_values_table']\
                    .at['commissioning_existing', 'Value']

    total_site_costs = pd.concat([total_site_costs, site_costs.sum(axis=1)])
    
    new_sites = [x.name for x in sites if x.dishes is None]

    new_site_costs = \
        pd.concat([new_site_costs, site_costs[site_costs.columns.intersection(new_sites)].sum(axis=1)])
    return total_site_costs, new_site_costs, site_costs

###
### Antenna Operations Costs Per Observation Year, which is primarily about staffing
###
def calculate_operations_costs(cost_config, sites, const):
    """
    Calculate the cost of operating the sites for a year, which is mostly about staffing.
    Uses both per-obeserving-night plus costs of some full-time staff.
    """
    array_stats = {}

    obs_per_year = cost_config.observations_per_year
    obs_days_per_year = cost_config.days_per_observation

    site_costs = pd.DataFrame(index=['Antenna operations'])

    operation_costs = pd.DataFrame(index=['total_non_local_labor_observation',
                                        'total_labor_needed_to_travel_observation',
                                        'total_local_labor_observation',
                                    #   'total_nonlocal_labor_monitoring',
                                    #   'total_local_labor_monitoring',
                                        'total_nonlocal_labor_maintenance',
                                        'total_local_labor_mainenance'])

    for site in sites:
        #
        # Antenna Operations costs
        # todo - don't we have operations costs for existing sites too?
        #
        siteindex = site.name

        operation_costs.loc[:, siteindex] = 0
        autonomy_scenario = cost_config.autonomy_of_operations
        site_region = site.region

        if not site.eht:
            # Antenna operations - total non-local labor during observations
            # eht_and_dedicated_obs_days_per_year = campaigns_per_year * campaign_duration

            labor_needed_to_travel = const['autonomy_mode_values_table']\
                .loc[autonomy_scenario, 'travel_campaign']
            remote_labor = const['autonomy_mode_values_table']\
                .loc[autonomy_scenario, 'remote_campaign_perday']

            # todo in the spreadsheet this is hardwired to NA labor, not by site region.
            na_remote_labor_cost_day = const['labor_cost_values_table'].at['science_salary',
                                                                        'N. America'] / 365

            total_non_local_labor_observation = obs_days_per_year * \
                (labor_needed_to_travel + remote_labor) * na_remote_labor_cost_day
            operation_costs.at['total_non_local_labor_observation',
                                siteindex] = total_non_local_labor_observation

            # Antenna operations - total travel during observations

            travel_cost = const['travel_cost_values_table']\
                .loc[:,site_region].loc['round_trip']
            per_diem = const['travel_cost_values_table']\
                .loc[:,site_region].loc['per_diem']

            operation_costs.at['total_labor_needed_to_travel_observation', siteindex] = \
                (labor_needed_to_travel * \
                    (obs_per_year * travel_cost) + (obs_days_per_year * per_diem))

            # Antenna operations - total local labor during observations
            local_labor = const['autonomy_mode_values_table']\
                .at[autonomy_scenario,'onsite_campaign']
            technician_cost_day = const['labor_cost_values_table']\
                .loc[:,site_region].loc['technician_salary'] / 365

            total_local_labor_observation = \
                obs_days_per_year * local_labor * technician_cost_day
            operation_costs\
                .at['total_local_labor_observation',siteindex] = total_local_labor_observation

            # # Antenna operations - total non-local labor during monitoring
            # nonlocal_labor_onsite_monitoring = const['autonomy_mode_values_table']\
            #     .loc[autonomy_scenario,'nonlocal_onsite_monitoring']
            # remote_labor_cost_day = const['labor_cost_values_table']\
            #     .at['science_salary',site_region] / 365
            # total_nonlocal_labor_monitoring = \
            #     monitoring_days_per_year * nonlocal_labor_onsite_monitoring * remote_labor_cost_day
            # operation_costs\
            #     .at['total_nonlocal_labor_monitoring',siteindex] = total_nonlocal_labor_monitoring

            # # Antenna operations - total local labor during monitoring
            # local_labor_onsite_monitoring = const['autonomy_mode_values_table']\
            #     .at[autonomy_scenario,'local_onsite_monitoring']

            # total_local_labor_monitoring = \
            #     monitoring_days_per_year * local_labor_onsite_monitoring * technician_cost_day
            # operation_costs\
            #     .at['total_local_labor_monitoring',siteindex] = total_local_labor_monitoring

            # Antenna operations - total non-local labor needed for maintenance
            nonlocal_labor_remote_maintenance = const['autonomy_mode_values_table']\
                .loc[autonomy_scenario,'nonlocal_remote_maint']
            scientist_cost_year = const['labor_cost_values_table']\
                .loc[:,site_region].loc['science_salary']

            total_nonlocal_labor_maintenance = \
                nonlocal_labor_remote_maintenance * scientist_cost_year
            operation_costs\
                .at['total_nonlocal_labor_maintenance',siteindex] = \
                    total_nonlocal_labor_maintenance
        else:
            # if it's an existing site, we just assume we have to pay $10k/night
            total_local_labor_observation = \
                obs_days_per_year * \
                    const['site_development_values_table']\
                    .at['existing_site_rental_per_night', 'Value']

            operation_costs\
                .at['total_local_labor_observation',siteindex] = total_local_labor_observation


        # Antenna operations - total local labor needed for maintenance
        local_labor_remote_maintenance = const['autonomy_mode_values_table']\
            .loc[autonomy_scenario,'local_onsite_maint']
        technician_cost_year = const['labor_cost_values_table']\
            .loc[:,site_region].loc['technician_salary']

        total_local_labor_maintenance = local_labor_remote_maintenance * technician_cost_year
        operation_costs\
            .at['total_local_labor_maintenance',siteindex] = total_local_labor_maintenance

        # add 'em up and plug into the site costs
        site_costs.at['Antenna operations',
                        siteindex] = operation_costs.loc[:, siteindex].sum()

    total_site_costs = site_costs.sum(axis=1)
    
    new_sites = [i for i,x in enumerate(sites) if x.dishes is None]
    new_site_costs = \
        site_costs[site_costs.columns.intersection(new_sites)].sum(axis=1)

    # assume we have a full-time staff of 10 people to coordinate operations
    cost_of_fulltime_staff = \
        3 * const['labor_cost_values_table'].loc[:,'N. America'].loc['science_salary'] + \
        4 * const['labor_cost_values_table'].loc[:,'N. America'].loc['engineering_salary'] + \
        3 * const['labor_cost_values_table'].loc[:,'N. America'].loc['technician_salary']

    total_site_costs['Fulltime staff'] = cost_of_fulltime_staff
    new_site_costs['Fulltime staff'] = cost_of_fulltime_staff

    return total_site_costs, new_site_costs, site_costs


##
## Data Management Costs
##
def calculate_data_costs(cost_config, sites_count, total_pb_per_year, collecting_days_per_year, \
    const):

    #
    # calculate data management costs
    #
    data_management_costs = pd.Series(dtype='float')
    # row to hold the category title
    data_management_costs['DATA MANAGEMENT'] = ''

    data_management_strategy = cost_config.data_management

    # data center capex
    data_management_costs['Cluster Build Cost'] = \
        const['data_management_option_values_table'].at[data_management_strategy, 'build_cost']

    # data center opex - personnel, data storage, computation costs

    # personnel costs
    data_management_costs['Personnel'] = \
        const['data_management_option_values_table']\
            .loc[data_management_strategy,'fte_required'] * \
                const['data_management_values_table'].at['fte_cost', 'Value']

    # costs to hold data in cold storage while we wait for it all to trickle in
    # if we're in Cloud mode, need to pay for 12 month minimum for cold storage
    months_to_hold = 12 if data_management_strategy == 'Cloud' else const[
        'data_management_values_table'].at['holding_months', 'Value']
    data_management_costs['Holding Data Storage Costs'] = \
        const['data_management_option_values_table']\
            .at[data_management_strategy,'holding_storage_perPB'] * \
                months_to_hold * \
                    total_pb_per_year

    # fast data storage cost, while we're processing
    data_management_costs['Fast Data Storage Costs'] = \
        const['data_management_option_values_table']\
            .loc[data_management_strategy,'fast_storage_perPB'] * \
                const['data_management_values_table']\
                    .at['processing_months', 'Value'] * total_pb_per_year

    # cost to transfer data from one class of storage to another
    data_management_costs['Transfer Costs'] = const['data_management_option_values_table']\
        .at[data_management_strategy,'data_xfer_perPB'] * total_pb_per_year

    # computation costs
    data_management_costs['Computation Costs'] = const['data_management_option_values_table']\
        .at[data_management_strategy, 'compute_fixed'] + \
            const['data_management_option_values_table']\
                .at[data_management_strategy,'compute_perPB'] * total_pb_per_year

    # station capex - recorders and media

    # cost for recorders - one per station
    data_management_costs['Site Recorders'] = sites_count * \
        const['data_management_values_table'].at['recorder_cost', 'Value'] * 1.2 # spares

    # media cost
    max_nights_media = const['data_management_values_table']\
        .at['media_on_hand', 'Value']
    nights_of_media = min(max_nights_media, collecting_days_per_year)
    total_pb_per_year_per_site = (total_pb_per_year / sites_count) if sites_count else 0
    pb_of_media_per_site = nights_of_media * (total_pb_per_year_per_site / collecting_days_per_year)
    data_management_costs['Site Media'] = sites_count * pb_of_media_per_site * \
        const['data_management_values_table'].at['media_cost_pb', 'Value']

    # station opex - cost to ship the data
    data_management_costs['Data Shipping'] = const['data_management_option_values_table']\
        .at[data_management_strategy,'shipping_perPB'] * total_pb_per_year

    return data_management_costs



"""
Copyright (c) 2022 Center for Astrophysics | Harvard & Smithsonian

This software is licensed for use as described in the LICENSE file in
the root directory of this distribution.

Testing the Target model

Originator: Aaron Oppenheimer March 2020
"""
from ngehtutil.target import Target
import unittest


class TestClass(unittest.TestCase):
    def test_target_list(self):
        a = Target.get_list()
        self.assertEqual(type(a),list)
        self.assertTrue(len(a)>0)


    def test_target_info(self):
        a = Target.get_list()
        info = Target.from_name(a[0])
        self.assertEqual(type(info),Target)

        with self.assertRaises(KeyError):
            info = Target.from_name('aaron')

    def test_default_target(self):
        t = Target.get_default()
        self.assertEqual(type(t),Target)




"""
Copyright (c) 2022 Center for Astrophysics | Harvard & Smithsonian

This software is licensed for use as described in the LICENSE file in
the root directory of this distribution.

Testing the Station model

Originator: Aaron Oppenheimer March 2020
"""
import unittest
from ngehtutil import Station, Array
from ngehtutil.station_weather import delete_sites

class TestClass(unittest.TestCase):
    """ tests for the Station """

    def test_station_list(self):
        """ show that we can get stations from the Station class """
        sl1 = Station.get_list()
        self.assertTrue(len(sl1)>0) # should get all stations

        array = Array.get_default_array_name()
        sl2 = Array.get_station_names(array)
        self.assertTrue(len(sl2)<=len(sl1))

        with self.assertRaises(KeyError):
            _ = Array.get_station_names('foo')

    def test_station_info(self):
        """ show that stations have names and things """
        station_list = Station.get_list()
        info = Station(station_list[0])
        self.assertEqual(type(info),Station)
        self.assertEqual(info.name, station_list[0])

    def test_station_diameter(self):
        """ show that we can get/set dish diameters """
        station = Station.from_name('CNI') # pick one we know doesn't have a dish
        self.assertEqual(station.dishes, None)

        station.set_diameter(10)
        self.assertEqual(len(station.dishes),1)
        self.assertEqual(station.dishes[0].diameter,10)

    def test_station_sefd(self):
        """ show that we get something resonable from SEFD """
        station_list = Station.get_list()
        station = Station(station_list[0])

        station.dishes = None
        with self.assertRaises(ValueError):
            sefd = station.SEFD(230,90)

        station.set_diameter(6)
        sefd = station.SEFD(230,90)
        self.assertTrue(sefd >= 0)

    def test_generic_station(self):
        """ show that we can make a generic station with defaults """
        station = Station('test')
        self.assertTrue(len(station.recording_frequencies),1)

    def test_station_frequencies(self):
        """ make sure the stations have between 1 and 3 frequencies """
        for _,stn in Station.get_all().items():
            num_freq = len(stn.recording_frequencies)
            self.assertTrue(num_freq >= 1)
            self.assertTrue(num_freq <= 3)

    def test_station_weather(self):
        """ show that we get weather data out of known stations """
        station = Station.from_name('HAY')
        data = station.get_weather('SEFD_info_230',2009,8,16)
        self.assertTrue(isinstance(data['data'], list))
        self.assertTrue(len(data['data']) > 0)
        self.assertTrue(isinstance(data['data'][0], tuple))
        delete_sites()

    def test_station_dupes(self):
        """ demonstrate that we can't have an array with stations with same name """
        stn1 = Station(name="s1")
        stn2 = Station(name="s2")
        stn3 = Station(name="s1")
        with self.assertRaises(ValueError):
            _ = Array(name="test", stations=[stn1, stn2, stn3])

    def test_station_immutable(self):
        """ make sure that when we request a station, we can't change it for other people """
        cni = Station.from_name("CNI")
        self.assertTrue(cni.dishes is None)
        cni.set_diameter(9)
        self.assertTrue(isinstance(cni.dishes, list))
        cni2 = Station.from_name("CNI")
        self.assertTrue(cni2.dishes is None)


"""
Copyright (c) 2022 Center for Astrophysics | Harvard & Smithsonian

This software is licensed for use as described in the LICENSE file in
the root directory of this distribution.

Testing Arrays

Originator: Aaron Oppenheimer March 2020
"""
from ngehtutil import *
import unittest

class TestClass(unittest.TestCase):
    """ tests for array objects """

    def test_array_list(self):
        """ show we can get arrays from our database """
        array = Array.get_list()
        self.assertEqual(type(array),list)
        self.assertTrue(len(array)>0)

    def test_default_array(self):
        """ show we can get a default array """
        def_array = Array.get_default_array_name()
        self.assertEqual(type(def_array),str)

        array = Array.get_default()
        self.assertEqual(type(array),Array)

    def test_array_init(self):
        """ show we can make an array from scratch """
        with self.assertRaises(ValueError):
            array = Array('test','test')

        with self.assertRaises(ValueError):
            array = Array('test',['test'])

    def test_array_immutable(self):
        """ show the arrays in the database can't be changed """
        arrays = Array.get_list()
        array1 = Array.from_name(arrays[0])
        array1_len = len(array1.stations())
        array2 = Array.from_name(arrays[1])
        array2_len = len(array2.stations())
        array2_len_orig = array2_len
        self.assertTrue(array1_len != array2_len)
        # we have two different arrays so let's now put the stations from 1 into 2
        array2.stations(array1.stations())
        array2_len = len(array2.stations())
        self.assertTrue(array1_len == array2_len)

        # now, make another array like array2 from the database and show that it's the old one
        array3 = Array.from_name(arrays[1])
        array3_len = len(array3.stations())
        self.assertTrue(array3_len == array2_len_orig)



"""
Copyright (c) 2022 Center for Astrophysics | Harvard & Smithsonian

This software is licensed for use as described in the LICENSE file in
the root directory of this distribution.

Code that exercises the utilities library. Generates information about how many dishes can be
built for fixed capital and operational costs to optimize a couple of performance parameters.

Originator: Aaron Oppenheimer March 2020
"""
from ngehtutil import Array, Target, Source, Schedule, Campaign, Program, Station
import random
import pandas as pd
import numpy as np
from copy import copy
import sys
import math


def make_array():
    # make a large array out of real stations
    array = Array.from_name('ngEHT Ref. Array 1.1A')
    stns = array.stations()
    # we want to have a few extras, just pick an average station and add a few
    s = Station.from_name('BAR')
    for i in range(0,4):
        s=copy(s)
        s.name=f'{s.name}{i}'
        stns.append(s)
    array.stations(stns)
    return array


def do_average(func):
    """
    Decorator that runs the function a number of times and returns the average
    """
    def wrapper(*args, **kwargs):
        d = pd.DataFrame()
        reps = 25
        for i in range(0,reps):
            d[i] = func(*args, **kwargs)
        return tuple(np.average(d,axis=1))
    return wrapper


@do_average
def number_for_diameter(diameter, construction_limit, ops_limit):
    """ 
    Find the number of antennas we can build that fit in the budget constraints. If an array
    is too expensive, drop one of the antennas at random.
    """
    t = Target.get_default()
    s = Source.get_default()
    sch = Schedule(obs_per_year=1, obs_days=5, obs_hours=24)
    c = Campaign(t,s,sch)

    array = make_array()
    while array.stations():
        cost = Program(array, c).calculate_costs(dish_size=diameter)
        if cost['TOTAL CAPEX'] <= construction_limit and \
            cost['ANNUAL OPEX'] <= ops_limit:
            # we found a set of dishes that are cheap enough
            break
        # too expensive - get rid of one site at random
        array._stations.pop(random.randrange(len(array._stations)))
    return len(array.stations()), cost['TOTAL CAPEX'], cost['ANNUAL OPEX']


def calc_metrics(construction_limit=999e9, ops_limit=999e9):

    data = pd.DataFrame()
    data.index.name = 'diameter'

    for d in range(4,16): # tdqm animates a progress bar
        num, c, o = number_for_diameter(diameter = d, \
           construction_limit=construction_limit, ops_limit = ops_limit )
        data.loc[d,'number'] = num
        sens = d * d
        data.loc[d,'sensitivity'] = sens
        fid = num * num
        data.loc[d,'fidelity'] = fid
        data.loc[d,'CAPEX'] = c
        data.loc[d,'OPEX'] = o
        data.loc[d,'TCO'] = c + (10 * o)

    # Create normalized version of the sensitivies and fidelities
    data.loc[:,'norm sens'] = data.loc[:,'sensitivity'] / np.max(data.loc[:,'sensitivity'])
    data.loc[:,'norm fid'] = data.loc[:,'fidelity'] / np.max(data.loc[:,'fidelity'])
    data.loc[:,'norm TCO'] = np.max(data.loc[:,'TCO']) - data.loc[:,'TCO']
    data.loc[:,'norm TCO'] = (data.loc[:,'norm TCO'] / np.max(data.loc[:,'norm TCO']))

    # reorder the columns so they match what's in the graphing spreadsheet
    data = data[[
        'number',
        'sensitivity',
        'fidelity',
        'OPEX',
        'CAPEX',
        'norm sens',
        'norm fid',
        'TCO',
        'norm TCO',
    ]]

    a = f'performance for const limit of {construction_limit} and ops limit of {ops_limit}\n'
    b = data.to_csv()
    return f'{a}\n{b}\n\n'


def main():
    output = []
    constlimit = 100e6
    opslimit_fraction = .10
    opslimit = constlimit * opslimit_fraction
    print(f'working on construction-limit only', file=sys.stderr)
    output.append(calc_metrics(construction_limit=constlimit))
    print(f'working on operations-limit only',  file=sys.stderr)
    output.append(calc_metrics(ops_limit=opslimit))
    print(f'working on both limits', file=sys.stderr)
    output.append(calc_metrics(construction_limit=constlimit, ops_limit=opslimit))

    # print the output in a format that is easily pasted into excel
    print('\n')
    print('\n\n\n\n\n\n\n\n'.join(output))

if __name__ == '__main__':
    main()


"""
Copyright (c) 2022 Center for Astrophysics | Harvard & Smithsonian

This software is licensed for use as described in the LICENSE file in
the root directory of this distribution.

Code to demonstrate use of the cost model

Originator: Aaron Oppenheimer March 2020
"""
from ngehtutil.cost import calculate_costs, CostConfig, get_cost_constants
from ngehtutil import *


def main():

    config = CostConfig(
        dish_size=6,
        observations_per_year=1,
        days_per_observation=60,
        hours_per_observation=60*8,
        no_upgrade=['ALMA', 'NOEMA', 'SMA']
    )

    array = Array.from_name('eht2022')
    stns = array.stations()
    new_stn_list = ['BOL','BRZ','PIKE','GAM','CAT','HAY','OVRO']
    new_stns = [Station.from_name(i) for i in new_stn_list]

    bima_stn_list = ['LAS', 'CNI', 'BAJA']
    bima_stns = [Station.from_name(i) for i in bima_stn_list]
    for s in bima_stns:
        s.set_diameter(6)

    stns = stns + new_stns + bima_stns

    costs = calculate_costs(config, stns)
    
    print(f'array: {array}')
    print('CONFIGURATION:')
    for x in dir(config):
        if not x[0]=='_':
            print(f'{x}: {getattr(config,x)}')
    for k,v in costs.items():
        print(f'{k}: {v}')


if __name__ == '__main__':
    main()



import ehtim as eh


files = [
'avery_m87_1_eofn',
'avery_sgra_eofn',
'jason_mad_eofn',
'roman_eofn',
'avery_m87_2_eofn',
'howes_m87',
'jason_mri_eofn',
'rowan_m87'
]

for f in files:
    im = eh.image.load_txt(f'models/{f}.txt')
    im.display(show=False, export_pdf=f"models/{f}.pdf")



import dash
from dash import dcc
from dash import html
from dash.dependencies import Input, Output
import plotly.express as px
import pandas as pd

data = pd.read_excel('exercise_out.xlsx', index_col=0)
data.loc['eht_sites'] = data.loc['eht_sites'].apply(lambda x: str(len(eval(x))))
data.loc['ngeht_sites'] = data.loc['ngeht_sites'].apply(lambda x: str(len(eval(x))))
data.loc['TOTAL CAPEX'] = data.loc['TOTAL CAPEX'].apply(lambda x: x/1e6)
data.loc['ANNUAL OPEX'] = data.loc['ANNUAL OPEX'].apply(lambda x: x/1e6)

app = dash.Dash(__name__)

inputs = []
factors = ['ngeht_sites', 'eht_sites', 'recording_frequencies', 'recording_bandwidth', 'dish_size', 'campaigns_per_year'] 

selectors = []
for i in factors:
    options = []
    for j in data.loc[i].unique():
        option = {'label':j, 'value': j}
        options.append(option)
    selector = html.Div([
        html.Label(i),
        dcc.Dropdown(
            id=f'{i}-dropdown',
            options=options,
        )],
        style={"width": "30%"}
    )
    selectors.append(selector)
    input = Input(f'{i}-dropdown', "value")
    inputs.append(input)


app.layout = html.Div(children=[
    html.H1(children='ngEHT Cost Estimating'),
    html.Div(children=selectors),
    # *selectors,
    dcc.Graph(
        id='example-graph'
    ),
])


@app.callback(
inputs=inputs,
output=Output('example-graph','figure')
)
def update_output(*args):
    x = data.copy()
    xmax = x.loc['TOTAL CAPEX'].max()+50
    ymax = x.loc['ANNUAL OPEX'].max()+1

    for i,f in enumerate(factors):
        if args[i]:
            x = x.loc[:,x.loc[f]==args[i]]

    fig = px.scatter(x.transpose(), x='TOTAL CAPEX', y='ANNUAL OPEX', \
        hover_data=factors,
        labels={
            'TOTAL CAPEX': 'Total Capex ($MM)',
            'ANNUAL OPEX': 'Annual Opex ($MM)',
        },
        title=f'Capex vs. Opex for {len(x.columns)} Configurations'
    )
    fig.update_layout(showlegend=False)
    fig.update_xaxes(range=[0, xmax])
    fig.update_yaxes(range=[0, ymax])

    return fig

if __name__ == '__main__':
    app.run_server(debug=True)



"""
Code to cost out array configurations

Executing this file runs a calculation on a test setup.

Requires 'costmodel' as a library accessible from the execution directory

AO 20211012
"""
import sys
from costmodel import CostConfig, calculate_costs, calculate_costs_over_time
from argparse import ArgumentParser
import logging
import pandas as pd
import os

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

def load_array_config(config_file, site_file):
    """ load the array configurations to cost out """
    logging.info('reading configurations from %s', config_file)

    # read configuration for new sites and convert it to a dict
    new_site_config = pd.read_excel(f'{CURRENT_DIR}/{config_file}',index_col=0, \
        sheet_name='NewSiteDefinition')
    new_sites = new_site_config.iloc[:,1:].to_dict()

    # read array definitions and convert to dict of list of sites
    all_site_info = pd.read_excel(f'{CURRENT_DIR}/{site_file}', index_col=0)
    array_configs = pd.read_excel(f'{CURRENT_DIR}/{config_file}',index_col=0, \
        sheet_name='ArrayDefinition')

    arrays = {}
    for array_config in array_configs.columns:
        array_sites = array_configs.loc[:,array_config].dropna().to_list()
        # if a site isn't in our site_info file, take it out of the list
        real_array_sites = []
        for site in array_sites:
            if site in all_site_info.columns:
                real_array_sites.append(site)
            else:
                logging.info(f"Don't know about site {site} so dropping it from {array_config}")
        arrays[array_config] = all_site_info.loc[:,real_array_sites]

    return new_sites, arrays

def write_output(output_file, output_data):
    """ save the data out """
    outfile = f'{CURRENT_DIR}/{output_file}'
    logging.info('writing data to %s',outfile)
    with pd.ExcelWriter(outfile) as writer: # pylint: disable=abstract-class-instantiated
        for d in output_data.items():
            d[1].to_excel(writer, sheet_name=d[0])

def main():
    """ do the cost calculations """

    parser=ArgumentParser(description='Calculate ngEHT array costs')
    parser.add_argument('--const_file', required=False, default=None,
        help='File containing all of the calculation constants (default uses module version)')
    parser.add_argument('--setup_file', required=False, default='array_setup.xlsx',
        help='File containing the array setup(s) to evaluate (default array_setup.xlsx)')
    parser.add_argument('--new_site_config', required=False, default=1,
        help='Which column in setup file defines the new site setup for per-array \
            analysis (default = 1)')
    parser.add_argument('--array', required=False, default=1,
        help='Which column in setup file defines the array for per-config analysis (default = 1)')
    parser.add_argument('--site_file', required=False, default='site_info.xlsx',
        help='File containing info about sites (default site_info.xlsx)')
    parser.add_argument('--output_file', required=False, default='cost_output.xlsx',
        help='Destination file for cost data (default cost_output.xlsx)')
    parser.add_argument('-v', '--verbose', required=False, action='store_true',
        help='Print some progress info')
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=0, format='%(message)s')

    output_data = {}

    new_site_configs, arrays = load_array_config(args.setup_file, args.site_file)

    # for each array config, assess it against the chosen new site config
    which_config = int(args.new_site_config)
    new_site_config = new_site_configs[list(new_site_configs.keys())[which_config-1]]
    the_config = CostConfig(**new_site_config)
    cost_data=pd.DataFrame()
    costs_over_time = {}
    for array_name, array_info in arrays.items():
        config_cost = calculate_costs(the_config, array_info, args.const_file)
        cost_data[array_name] = pd.Series(new_site_config).append(config_cost)
        config_cost_over_time = calculate_costs_over_time(new_site_config, array_name, config_cost)
        costs_over_time[array_name] = config_cost_over_time
    output_data['Costs by array'] = cost_data
    output_data['Costs over time by array'] = pd.concat(costs_over_time, axis=1)

    # for each site config, assess it against the chosen array
    which_array = int(args.array)
    array_name = list(arrays.keys())[which_array-1]
    array = arrays[array_name]
    cost_data=pd.DataFrame()
    costs_over_time = {}
    for config in new_site_configs.items():
        the_config = CostConfig(**config[1])
        config_cost = calculate_costs(the_config, array, args.const_file)
        config[1]['Array'] = array_name # stick the name of the array into the output
        cost_data[config[0]] = pd.Series(config[1]).append(config_cost)
        config_cost_over_time = calculate_costs_over_time(config[1], array_name, config_cost)
        costs_over_time[config[0]] = config_cost_over_time
    output_data['Costs by config'] = cost_data
    output_data['Costs over time by config'] = pd.concat(costs_over_time, axis=1)

    write_output(args.output_file, output_data)


if __name__ == '__main__':
    main()



"""
Copyright (c) 2022 Center for Astrophysics | Harvard & Smithsonian

This software is licensed for use as described in the LICENSE file in
the root directory of this distribution.

Testing the cost model

Originator: Aaron Oppenheimer March 2020
"""
import unittest
from pandas import Series
from ngehtutil.cost import calculate_costs, get_cost_constants
from ngehtutil.cost import CostConfig
from ngehtutil.cost.cost_model import *
from ngehtutil import *
from copy import copy

class CostTestClass(unittest.TestCase):
    """ A set of tests for the cost module """

    def test_costmodel_constants(self):
        """
        In some circumstances we want a copy of the cost constants dictionary. Make sure
        we get a deep copy, not just a reference to the one loaded by the module.
        """
        a = get_cost_constants()
        b = a
        c = get_cost_constants()
        self.assertIs(a, b)
        self.assertIsNot(a,c)

        t = list(a.keys())[0]
        self.assertIs(a[t],b[t])
        self.assertIsNot(a[t],c[t])

    def test_costmodel_stationobjects(self):
        config = CostConfig()
        array = Array.from_name(Array.get_list()[0])
        costs, site_costs = calculate_costs(config, array.stations())
        self.assertEqual(type(costs), dict)
        self.assertEqual(type(site_costs), dict)

    def test_capital_costs(self):
        config = CostConfig()
        array = Array.from_name(Array.get_list()[0])
        total_site_costs, new_site_costs, site_costs = calculate_capital_costs(config, array.stations(), \
            get_cost_constants())
        all_site = sum([x for x in total_site_costs.to_dict().values() if not type(x) is str])
        all_new = sum([x for x in new_site_costs.to_dict().values() if not type(x) is str])
        self.assertTrue(all_site >= all_new)

        # make sure all of the values in the costs are real
        for k,v in total_site_costs.items():
            if type(v) is int:
                self.assertFalse(v<0)

    def test_cost_config(self):
        config = CostConfig()
        array = Array.from_name(Array.get_list()[0])
        config.recording_frequencies = 2
        total_site_costs1, _, _ = calculate_capital_costs(config, array.stations(), \
            get_cost_constants())

        config.recording_frequencies = 3
        total_site_costs2, _, _ = calculate_capital_costs(config, array.stations(), \
            get_cost_constants())

        self.assertTrue(total_site_costs1['Receiver and Backend costs'] < \
            total_site_costs2['Receiver and Backend costs'])

    def test_operations_costs(self):
        config = CostConfig()
        array = Array.from_name(Array.get_list()[0])
        total_site_costs, new_site_costs, per_site_costs = \
            calculate_operations_costs(config, array.stations(), get_cost_constants())
        all_site = sum([x for x in total_site_costs.to_dict().values() if not type(x) is str])
        all_new = sum([x for x in new_site_costs.to_dict().values() if not type(x) is str])
        self.assertTrue(all_site >= all_new)

    def test_data_costs(self):
        """
        TODO make this do something useful
        """
        config = CostConfig()
        array = Array.from_name(Array.get_list()[0])
        data_costs = calculate_data_costs(config, len(array.stations()), 1, 10, \
            get_cost_constants())
        self.assertEqual(type(data_costs.to_dict()), dict)

    def test_station_copies(self):
        """
        Verify that an array made of multiples of the same station works properly
        """

        config = CostConfig()

        # get the costs of an empty array
        array0 = Array('test',[])
        costs0, _ = calculate_costs(config, array0.stations())

        stn = Station.from_name('LOS')
        array1 = Array('test',[stn]) # one LOS
        costs1, _ = calculate_costs(config, array1.stations())

        multi = 10
        stns2 = [copy(stn) for _ in range(0,multi)]
        for i,a in enumerate(stns2):
            a.name = f'{a.name}_{i}'
        array2 = Array('test',stns2) # multiple LOSs
        costs2, _ = calculate_costs(config, array2.stations())

        base_cost = costs0['TOTAL CAPEX']
        array1_cost = costs1['TOTAL CAPEX'] - base_cost
        array2_cost = costs2['TOTAL CAPEX'] - base_cost
        array2_cost_per_site = array2_cost / multi

        # an array of 1 and an array of 10 should cost close to 10x - not exactly due to rounding
        # in the amount of data grabbed per site
        self.assertTrue(math.isclose(array1_cost, array2_cost_per_site, rel_tol=0.02))

    def test_dish_const_cost(self):
        # verify that costs are different for a site where we have to build a dish vs. one where
        # the dish already exists
        config = CostConfig()

        stn1 = Station.from_name('HAY') # pick one that already has a dish
        self.assertTrue(stn1.dishes is not None)
        cost1, _ = calculate_costs(config, [stn1])

        stn2 = copy(stn1)
        stn2.dishes = None # get rid of the dish
        self.assertTrue(stn2.dishes is None)
        cost2, _ = calculate_costs(config, [stn2])

        print(cost1)
        print('\n\n')
        print(cost2)

        self.assertTrue(cost1['TOTAL CAPEX'] < cost2['TOTAL CAPEX'])


    def test_design_NRE_cost(self):
        # verify that we get a design NRE cost
        config = CostConfig()

        stn1 = Station.from_name('HAY') # pick one that already has a dish
        cost1, _ = calculate_costs(config, [stn1])

        stn2 = copy(stn1)
        stn2.dishes = None # get rid of the dish
        cost2, _ = calculate_costs(config, [stn2])

        self.assertTrue(cost1['DESIGN NRE'] < cost2['DESIGN NRE'])



"""
Classes to describe VLBI program
"""
from lib2to3.pytree import type_repr
from ngehtutil import calculate_costs, CostConfig, Campaign, Array

class Program:
    array = None
    campaign = None

    def __init__(self, array=None, campaign=None):
        if type(array) is not Array:
            raise TypeError
        self.array = array

        if type(campaign) is not Campaign:
            raise TypeError
        self.campaign = campaign

    def calculate_costs(self, **kwargs):
        """ Use the cost model to figure out what an array and campaigns cost. """
        """ Pass in CostConfig attributes to override them. """

        if type(self.array) is not Array:
            raise ValueError("Program not configured with an Array")

        if not self.campaign:
            raise ValueError("Program not configured with Campaigns")

        config = CostConfig()
        config.observations_per_year = self.campaign.schedule.obs_per_year
        config.days_per_observation = self.campaign.schedule.obs_days
        config.hours_per_observation = self.campaign.schedule.obs_hours
        config.recording_frequencies = 1

        for k,v in kwargs.items():
            setattr(config, k, v)

        costs, site_costs = calculate_costs(config, self.array.stations())
        return costs, site_costs




"""
Code to cost out array configurations for science meeting exercise

AO 20211021
"""
import itertools
import pandas as pd
from tqdm import tqdm
import sys
sys.path.insert(0,'..')
from cost_config import CostConfig
from cost_model import calculate_costs
import os

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

# options for each of the input parameters
DISH_DIAMETERS = [4, 6, 8, 10]
NGEHT_ANTENNAS = [ # 4, 6, 8, 10 sites
    [],
    ['BAJA'],
    ['OVRO', 'LLA', 'BAJA', 'HAY'],
    ['OVRO', 'LLA', 'BAJA', 'HAY', 'CNI', 'GAM'],
    ['OVRO', 'LLA', 'BAJA', 'HAY', 'CNI', 'GAM', 'GARS', 'CAT'],
    ['OVRO', 'LLA', 'BAJA', 'HAY', 'CNI', 'GAM', 'GARS', 'CAT', 'SGO', 'NZ']
]
EHT_SITES = [ # all 11, 4
    [],
    ['GLT', 'SMA', 'LMT'],
    ['ALMA', 'APEX', 'GLT', 'IRAM-30m', 'JCMT', 'KP', 'LMT', 'NOEMA', 'SMA', 'SMT', 'SPT'],
]
BANDWIDTH = [4, 8, 12]
FREQUENCIES = [2, 3]
CAMPAIGNS = [1, 4, 12]
PARAMETER_NAMES = [
    'dish_size', 'ngeht_sites', 'eht_sites', 'recording_bandwidth', 'recording_frequencies', \
    'campaigns_per_year'
]

ALL_SITE_INFO = None
def load_array_config(sites, site_file='site_info.xlsx'):
    """ load the site information """
    global ALL_SITE_INFO
    if ALL_SITE_INFO is None: # just read it once
        ALL_SITE_INFO = pd.read_excel(f'{CURRENT_DIR}/{site_file}', index_col=0)
    return ALL_SITE_INFO.loc[:,sites]


def main():
    """ do the cost calculations """

    # make all possible combinations and turn it into a list of dicts describing each of the
    # parameters as needed by the cost model
    all_combinations = list(itertools.product(
        DISH_DIAMETERS,
        NGEHT_ANTENNAS,
        EHT_SITES,
        BANDWIDTH,
        FREQUENCIES,
        CAMPAIGNS,
    ))
    all_configurations = [CostConfig(**dict(zip(PARAMETER_NAMES, c))) for c in all_combinations]

    output_data=[]
    for configuration in tqdm(all_configurations):
        array = load_array_config(configuration.eht_sites + configuration.ngeht_sites)
        config_cost = calculate_costs(configuration, array)
        output_data.append(pd.Series(configuration).append(config_cost))

    pd.DataFrame(output_data).transpose().to_excel(f'{CURRENT_DIR}/exercise_out.xlsx')

if __name__ == '__main__':
    main()
